
from . import constants
from . import anomaly
import numpy as np

def current_radius(r_p, e, theta):
  """
  Calculate current radius from perihelion radius, eccentricity, and true anomaly.
  """
  r = r_p*(1 + e) / (1 + e * np.cos(theta))
  return r

class Orbit:
    def __init__(self, peri_r = 0, tp = 0, e = 0):
        """
        Contains orbital parameters.
        Args:
            e (float): Eccentricity of the orbital (0-1).
            peri_r (float): Perihelion radius.
            tp (float): Period time.
        """
        self.e = e # Eccentricity (0-1)
        self.peri_r = peri_r # Radius of perihelion (initial position)
        self.tp = tp # Period time

class Body:
  def __init__(self, name, color, mass, mean_diameter, orbit = None):
        """
        Initializes a Body object for 2D simulation.
        Args:
            name (str): Name of the body.
            mass (float): Mass of the body (kg).
            mean_diameter (float): Mean diameter of the body (m).
            color (str): Color for plotting.
            orbit (Orbit): Orbital parameters.
        """
        self.name = name
        self.mass = mass
        self.mean_diameter = mean_diameter
        self.color = color
        self.orbit = orbit
        if orbit is not None:
          self.x = orbit.peri_r # Initial x-coordinate
        else:
          self.x = 0.0  # Initial x-coordinate
        self.y = 0.0  # Initial y-coordinate

  def update(self, t, c_x = 0.0 , c_y = 0.0, logarithmic = False, depth = 1):
        # Check if orbit exists before trying to use it
        if self.orbit is None:
             self.x = c_x
             self.y = c_y
             return

        mean_anomaly = t * 2 * np.pi / self.orbit.tp # Mean anomaly is always 0 at t=0 (-> no addition of it here)
        true_anomaly = anomaly.Anomaly(mean_anomaly, type = 'M').theta(self.orbit.e)
        r = current_radius(self.orbit.peri_r,self.orbit.e, true_anomaly)
        if logarithmic:
            r = np.log10(r) / depth

        self.x = r * np.cos(true_anomaly) # Local x coordinate
        self.x = self.x + c_x # Global x coordinate
        self.y = r * np.sin(true_anomaly)    # Local y coordinate
        self.y = self.y + c_y # Global y coordinate

  def get_bodies(self):
        return [self]

class System:
    _instances = []
    def __init__(self, name, center, orbiting=None):
        """
        Initializes an orbital System.

        Args:
            center (Body): The central body of the system.
            bodies (list of Body, optional): List of bodies orbiting the center. Defaults to None.
        """
        self.__class__._instances.append(self)
        self.name = name
        self.center = center
        # self.orbiting_objects = orbiting if orbiting is not None else []
        if orbiting is not None:
            if isinstance(orbiting, (System, Body)):
                self.orbiting_objects= [orbiting]
            if isinstance(orbiting, (list, tuple)) and all(isinstance(item, (System, Body)) for item in orbiting):
                self.orbiting_objects = orbiting
        else:
            self.orbiting_objects = []

#        self.all_bodies = [center]

        # for body in self.orbiting_objects:
        #     self.all_bodies.append(body.get_bodies())
    @classmethod
    def get_all_instances(cls):
        # Return the list of all instances
        return cls._instances

    def add_body_or_system(self, body):
        self.orbiting_objects.append(body)

    # def add_system(self, body):
    #     self.bodies.append(body)

    def give_center_body(self, center_body):
        self.center = center_body

    def set_center_pos(self, coords):
        self.center.x = coords[0]
        self.center.y = coords[1]
    '''
    def update(self, t, c_x, c_y, logarithmic = False, depth = 0):
        if self.center.orbit:
          self.center.update(t, c_x, c_y, logarithmic)
        else:
          self.center.x = c_x
          self.center.y = c_y
        depth = depth + 1
        for obj in self.orbiting_objects:
          obj.update(t, self.center.x, self.center.y, logarithmic, depth)
    '''

    def update(self, t, c_x, c_y, logarithmic=False, depth=0):
        # --- MODIFIED LOGIC for positioning the center ---
        # Case 1: This system is a TOP-LEVEL system called by Universe (c_x, c_y are 0,0)
        #         AND its center has an orbit defined (e.g., Earth in EarthMoon sim).
        #         Force the center to stay at (0,0) for this context.
        if depth == 0 and self.center.orbit:
            self.center.x = 0.0
            self.center.y = 0.0
            # Do NOT call self.center.update() in this specific case.

        # Case 2: The center body has an orbit defined AND it's NOT the specific case above
        #         (i.e., this system is orbiting an outer center where c_x, c_y != 0,0).
        #         Update the center's position based on its orbit relative to c_x, c_y.
        elif self.center.orbit:
            # Update center relative to the outer center c_x, c_y
            # NOTE: Pass depth to Body.update if it uses it (it doesn't currently, but good practice)
            self.center.update(t, c_x, c_y, logarithmic, depth)

        # Case 3: The center body has NO orbit defined.
        #         Place it exactly at the reference point c_x, c_y.
        else:
            self.center.x = c_x
            self.center.y = c_y
        # --- END MODIFIED LOGIC ---

        # Now update all orbiting objects relative to the *final* position of the center
        orbiting_center_x = self.center.x
        orbiting_center_y = self.center.y
        depth = depth + 1  # Increment depth for subsystems
        for obj in self.orbiting_objects:
            # Pass the calculated center position and incremented depth
            obj.update(t, orbiting_center_x, orbiting_center_y, logarithmic, depth)

    def get_bodies(self):
       # return [self.center].append(self.orbiting_objects)
       bodies = [self.center]
       for obj in self.orbiting_objects:
           bodies.extend(obj.get_bodies())
       return bodies

class Universe:
    def __init__(self, systems=None):
        """
        Initializes simulation space.

        Args:
            systems : List of systems in the simulation.
        """

        # if systems is None:
        #     systems = []
        if systems is not None:
            if isinstance(systems, (System, Body)):
                systems = [systems]
            if isinstance(systems, (list, tuple)) and all(isinstance(item, (System, Body)) for item in systems):
                systems = systems
        else:
            systems = []

        self.all_elements = systems
        self.all_bodies = []
        for elem in self.all_elements:
            self.all_bodies.extend(elem.get_bodies())

    def update(self, time, logarithmic = False):
        for elem in self.all_elements:
          elem.update(time, 0.0, 0.0, logarithmic)

    def get_bodies(self):
        return self.all_bodies

Sun = Body(
    name = "Sun",
    color = "yellow",
    mass = constants.Sun_m,
    mean_diameter = constants.Sun_rm)

Earth = Body(
    name = "Earth",
    color = "blue",
    mass = constants.Earth_m,
    mean_diameter = constants.Earth_rm,
    orbit = Orbit(constants.Earth_perihelion, constants.Earth_T, constants.Earth_e))

Moon = Body(
    name = "Moon",
    color = "grey",
    mass = constants.Moon_m,
    mean_diameter = constants.Moon_rm,
    orbit = Orbit(constants.Moon_perihelion, constants.Moon_T, constants.Moon_e))

Mercury = Body(
    name="Mercury",
    color="darkgrey",
    mass=constants.Mercury_m,
    mean_diameter=constants.Mercury_rm,
    orbit=Orbit(constants.Mercury_perihelion, constants.Mercury_T, constants.Mercury_e)
)

Venus = Body(
    name="Venus",
    color="goldenrod",
    mass=constants.Venus_m,
    mean_diameter=constants.Venus_rm,
    orbit=Orbit(constants.Venus_perihelion, constants.Venus_T, constants.Venus_e)
)

Mars = Body(
    name="Mars",
    color="red",
    mass=constants.Mars_m,
    mean_diameter=constants.Mars_rm,
    orbit=Orbit(constants.Mars_perihelion, constants.Mars_T, constants.Mars_e)
)

Jupiter = Body(
    name="Jupiter",
    color="orange",
    mass=constants.Jupiter_m,
    mean_diameter=constants.Jupiter_rm,
    orbit=Orbit(constants.Jupiter_perihelion, constants.Jupiter_T, constants.Jupiter_e)
)

Saturn = Body(
    name="Saturn",
    color="khaki",
    mass=constants.Saturn_m,
    mean_diameter=constants.Saturn_rm,
    orbit=Orbit(constants.Saturn_perihelion, constants.Saturn_T, constants.Saturn_e)
)

Uranus = Body(
    name="Uranus",
    color="lightblue",
    mass=constants.Uranus_m,
    mean_diameter=constants.Uranus_rm,
    orbit=Orbit(constants.Uranus_perihelion, constants.Uranus_T, constants.Uranus_e)
)

Neptune = Body(
    name="Neptune",
    color="deepskyblue",
    mass=constants.Neptune_m,
    mean_diameter=constants.Neptune_rm,
    orbit=Orbit(constants.Neptune_perihelion, constants.Neptune_T, constants.Neptune_e)
)

Pluto = Body(
    name="Pluto",
    color="lightgrey",
    mass=constants.Pluto_m,
    mean_diameter=constants.Pluto_rm,
    orbit=Orbit(constants.Pluto_perihelion, constants.Pluto_T, constants.Pluto_e)
)

Eris = Body(
    name="Eris",
    color="white",
    mass=constants.Eris_m,
    mean_diameter=constants.Eris_rm,
    orbit=Orbit(constants.Eris_perihelion, constants.Eris_T, constants.Eris_e)
)

Moon2 = Body(
    name = "Moon2",
    color = "grey",
    mass = constants.Moon_m,
    mean_diameter = constants.Moon_rm,
    orbit = Orbit(constants.Moon_perihelion * 150, constants.Moon_T *2, constants.Moon_e * 0.2))

Moon3 = Body(
    name = "Moon3",
    color = "grey",
    mass = constants.Moon_m,
    mean_diameter = constants.Moon_rm * 0.6,
    orbit = Orbit(constants.Moon_perihelion * 50, constants.Moon_T, constants.Moon_e))

EarthMoon = System(name = "EarthMoon", center = Earth, orbiting = Moon)

SunEarthMoon = System(name="SunEarthMoon", center = Sun, orbiting = EarthMoon)

Solar_system = System(name="Solar_system", center = Sun, orbiting = [
    Mercury,
    Venus,
    EarthMoon,
    Mars,
    Jupiter,
    Saturn,
    Uranus,
    Neptune,
    Pluto,
    Eris
])

Solar_system0 = System(name="Solar_system0", center = Sun, orbiting = [
    Mercury,
    Venus,
    EarthMoon,
    Mars,
])

Dwarfs = System(name="Dwarfs", center = Sun, orbiting = [Pluto, Eris])


import warnings
from copy import copy

from numpy import allclose as almost_equal , divmod
from scipy.constants import pi

from . import orbital_conversions as oc
from .orbital_conversions import (
    elements_4_apsides, radius_from_alt, alt_from_radius, impulse_from_finite, saved_state)

class ReprHelperMixin:
    """Provides a simplified __repr__ for classes to assist with debugging and logging."""
    def __repr__(self):
        return f"<{self.__class__.__name__}>"

from . import anomaly as anom
from .anomaly import (
     mean_anomaly_from_eccentric, mean_anomaly_from_true)   


########################################################################
#ez most máshonnan veszem 
########################################################################
_copy = copy

class Operation(object):
    """Abstract base class for orbital operations that can be applied, plotted, or combined."""

    def plot(self, orbit, plotter, next_operation=None):
        """Plot the operation on the given orbit using the provided plotter."""
        if hasattr(self, '__plot__') and callable(getattr(self, '__plot__')):
            self.__plot__(orbit, plotter, next_operation)

    def __add__(self, other):
        """Add two operations together."""
        if isinstance(other, Operation):
            return Maneuver(self, other)
        else:
            return NotImplementedError()
        
    def __radd__(self, other):
        """Right add two operations together."""
        if isinstance(other, Operation):
            return Maneuver(other, self)
        else:
            return NotImplementedError()

########################################################################
########################################################################
        
class ImpulseOperation(Operation):
    """Base class for instantaneous velocity changes applied to an orbit (delta-v maneuvers)."""

    def __init__(self):
        super(ImpulseOperation, self).__init__()

    def velocity(self):
        """Return the velocity of the impulse operation."""
        raise NotImplementedError()
    
########################################################################
########################################################################
    
class TimeOperation(Operation):
    """Base class for time-based operations that propagate the orbit forward in time."""

    def __init__(self, time):
        super(TimeOperation, self).__init__()
        self.time = time

    def time(self):
        """Return the time of the operation."""
        raise NotImplementedError()
    
########################################################################
#APOCENTER
########################################################################


class SetApocenterRadiusTo(ReprHelperMixin, ImpulseOperation): 
    """
    Set the apocenter radius of the orbit to a specified value.

    This operation should be applied at the pericenter.

    Parameters
    ----------
    r : float
        The desired apocenter radius (in meters).
    """

    def __init__(self, apocenter_radius, **kwargs):
        super(SetApocenterRadiusTo, self).__init__()
        self.apocenter_radius = apocenter_radius
        #self.kwargs = kwargs

    def __repr__(self):
        return f"SetApocenterRadiusTo(radius={self.radius}, kwargs={self.kwargs})"
    
    def __apply__(self, orbit):
        a , e = elements_4_apsides(self.apocenter_radius, orbit.pericenter_radius)

        orbit.a = a
        orbit.e = e

        orbit.v = orbit.v

    def __plot__(self, orbit, plotter, next_operation=None):
        if orbit.apocenter_radius > self.apocenter_radius:
            label = 'Apocenter radius decreased'
        else:
            label = 'Apocenter radius increased'
        self.__apply__(orbit)
       
        """Plot the operation on the given orbit using the provided plotter."""
        with saved_state(orbit):
            if (next_operation is not None and
                isinstance(next_operation, TimeOperation)):
                orbit.apply_maneuver(next_operation)
                f2 = orbit.theta
                f2 = 2 * pi

        plotter.plot_apsides(orbit ,f1=0,  f2=f2 , label=label)


    def velocity(self, orbit):
        with saved_state(orbit):
            # get veloctiy and pericenter
            orbit.propagate_anomaly_to(M=0)
            old_v = orbit.v

            a, e = elements_4_apsides(self.apocenter_radius, orbit.pericenter_radius)

            orbit.semi_major_axis = a
            orbit.eccentricity = e

            new_v = orbit.v

        return new_v - old_v
      

    def represnetation_helper(self , r):
        r.positional_from_attr('apocenter_atlitude')

########################################################################
########################################################################

class SetApocenterAltitudeTo(ReprHelperMixin, ImpulseOperation):
    '''Set  apocenter alitude to a given value, when used must be at orbital pericenter'''

    def __init__(self, apocenter_altitude):
        super(SetApocenterAltitudeTo, self).__init__()
        self.apocenter_radius = apocenter_altitude

    def __apply__(self, orbit):
        apocenter_radius = orbit.body.mean_radius + self.apocenter_radius
        a, e = elements_4_apsides(radius_from_alt(self.apocenter_radius), orbit.pericenter_radius)

        orbit.a = a
        orbit.e = e

        orbit.v = orbit.v

    def __plot__(self, orbit, plotter, next_operation=None):
        radius = radius_from_alt(self.apocenter_radius, orbit.body)
        if orbit.apocenter_radius > radius:
            label = 'Apocenter radius decreased'
        else:
            label = 'Apocenter radius increased'
        self.__apply__(orbit)

        with saved_state(orbit):
            if (next_operation is not None and
                isinstance(next_operation, TimeOperation)):
                orbit.apply_maneuver(next_operation)
                f2 = orbit.theta
                if f2  == 0:
                    f2 = 2 * pi
                else:
                    f2 = 2 * pi

        plotter._plot_apsides(orbit ,f1=0,  f2=f2 , label=label)

    def velocity(self, orbit):
        with saved_state(orbit):
            orbit.propagate_anomaly_to(M=0)
            old_v = orbit.v

            a, e = elements_4_apsides(radius_from_alt(self.apocenter_radius), orbit.pericenter_radius)

            orbit.a = a
            orbit.e = e

            new_v = orbit.v

        return new_v - old_v
    
    def __repr__(self, r):
        r.positional_from_attr('apocenter_altitude')

########################################################################
########################################################################

class ChangeApocenterBy(ReprHelperMixin, ImpulseOperation): 
    '''Operation for changing apocenter radius, when applied must be at orpbital pericenter.'''
    
    def __init__(self, delta):
        super(ChangeApocenterBy, self).__init__()
        self.delta = delta

    def __apply__(self, orbit):
        a, e = elements_4_apsides(orbit.apocenter_radius + self.delta, orbit.pericenter_radius)
    
        orbit.a = a
        orbit.e = e

        orbit.v = orbit.v

    def __plot__(self, orbit, plotter, next_operation=None):
        if self.delta > 0:
            label = 'Apocenter radius increased'
        else:
            label = 'Apocenter radius decreased'
        self.__apply__(orbit)

        with saved_state(orbit):
            if (next_operation is not None and
                isinstance(next_operation, TimeOperation)):
                orbit.apply_maneuver(next_operation)
                f2 = orbit.theta
                if f2  == 0:
                    f2 = 2 * pi
                else:
                    f2 = 2 * pi

        plotter._plot_apsides(orbit ,f1=0,  f2=f2 , label=label)

    def velocity(self, orbit):
        with saved_state(orbit):
            # get veloctiy and pericenter
            orbit.propagate_anomaly_to(M=0)
            old_v = orbit.v

            a, e = elements_4_apsides(orbit.apocenter_radius + self.delta, orbit.pericenter_radius)

            orbit.a = a
            orbit.e = e

            new_v = orbit.v

        return new_v - old_v

    def __repr__(self, r):
        r.positional_from_attr('delta')

########################################################################
#PERICENTER
########################################################################

class SetPericenterRadiusTo(ReprHelperMixin, ImpulseOperation):
    """
    Set the pericenter radius of the orbit to a specified value.

    This operation should be applied at the apocenter.

    Parameters
    ----------
    r : float
        The desired pericenter radius (in meters).
    """


    def __init__(self, pericenter_radius):
        super(SetPericenterRadiusTo, self).__init__()
        self.pericenter_radius = pericenter_radius

    def __apply__(self, orbit):
        a, e = elements_4_apsides(orbit.apocenter_radius, self.pericenter_radius)

        orbit.a = a
        orbit.e = e

        orbit.v = orbit.v

    def __plot__(self, orbit, plotter, next_operation=None):
        if orbit.pericenter_radius > self.pericenter_radius:
            label = 'Pericenter radius decreased'
        else:
            label = 'Pericenter radius increased'
        self.__apply__(orbit)

        with saved_state(orbit):
            if (next_operation is not None and
                isinstance(next_operation, TimeOperation)):
                orbit.apply_maneuver(next_operation)
                f2 = orbit.theta
                if f2  == 0:
                    f2 = 2 * pi
                else:
                    f2 = 2 * pi

        plotter._plot_apsides(orbit ,f1=0,  f2=f2 , label=label)


    def velocity(self, orbit):
        with saved_state(orbit):
            orbit.propagate_anomaly_to(M=0)
            old_v = orbit.v

            a, e = elements_4_apsides(self.pericenter_radius, orbit.apocenter_radius)

            orbit.a = a
            orbit.e = e

            new_v = orbit.v

        return new_v - old_v
    
    def __repr__(self, r):
        r.positional_from_attr('pericenter_radius')

########################################################################
########################################################################

class SetPericenterAltitudeTo(ReprHelperMixin, ImpulseOperation):
    '''Set pericenter alitude to a given value, when used must be at orbital apocenter'''

    def __init__(self, pericenter_altitude):
        super(SetPericenterAltitudeTo, self).__init__()
        self.pericenter_radius = pericenter_altitude

    def __apply__(self, orbit):
        pericenter_radius = orbit.body.mean_radius + self.pericenter_altitude
        a, e = elements_4_apsides(orbit.apocenter_radius, radius_from_alt(self.pericenter_radius))

        orbit.a = a
        orbit.e = e

        orbit.v = orbit.v

    def __plot__(self, orbit, plotter, next_operation=None):
        radius = radius_from_alt(self.pericenter_radius, orbit.body)
        if orbit.pericenter_radius > radius:
            label = 'Pericenter radius decreased'
        else:
            label = 'Pericenter radius increased'
        self.__apply__(orbit)

        with saved_state(orbit):
            if (next_operation is not None and
                isinstance(next_operation, TimeOperation)):
                orbit.apply_maneuver(next_operation)
                f2 = orbit.theta
                if f2  == 0:
                    f2 = 2 * pi
                else:
                    f2 = 2 * pi

        plotter._plot_apsides(orbit ,f1=0,  f2=f2 , label=label)

    def velocity(self, orbit):
        with saved_state(orbit):
            # get veloctiy and pericenter
            orbit.propagate_anomaly_to(M=0)
            old_v = orbit.v

            a, e = elements_4_apsides(radius_from_alt(self.pericenter_radius), orbit.apocenter_radius)

            orbit.a = a
            orbit.e = e

            new_v = orbit.v

        return new_v - old_v
    
    def __repr__(self, r):
        r.positional_from_attr('pericenter_altitude')


class ChangePericenterBy(ReprHelperMixin, ImpulseOperation):
    '''Operation for changing pericenter radius, when applied must be at orbital apocenter.'''
    
    def __init__(self, delta):
        super(ChangePericenterBy, self).__init__()
        self.delta = delta

    def __apply__(self, orbit):
        a, e = elements_4_apsides(orbit.apocenter_radius, orbit.pericenter_radius + self.delta)

        orbit.a = a
        orbit.e = e

        orbit.v = orbit.v

    def __plot__(self, orbit, plotter, next_operation=None):
        if self.delta > 0:
            label = 'Pericenter radius increased'
        else:
            label = 'Pericenter radius decreased'
        self.__apply__(orbit)

        with saved_state(orbit):
            if (next_operation is not None and
                isinstance(next_operation, TimeOperation)):
                orbit.apply_maneuver(next_operation)
                f2 = orbit.theta
                if f2  == 0:
                    f2 = 2 * pi
                else:
                    f2 = 2 * pi

        plotter._plot_apsides(orbit ,f1=0,  f2=f2 , label=label)

    def velocity(self, orbit):
        with saved_state(orbit):
            # get veloctiy and pericenter
            orbit.propagate_anomaly_to(M=0)
            old_v = orbit.v

            a, e = elements_4_apsides(orbit.apocenter_radius, orbit.pericenter_radius + self.delta)

            orbit.a = a
            orbit.e = e

            new_v = orbit.v

        return new_v - old_v
    
    def __repr__(self, r):
        r.positional_from_attr('delta')

########################################################################
#INCLANATION
########################################################################

class SetInclinationTo(ReprHelperMixin, ImpulseOperation):
    '''Set inclination to a given value, when used must be at ascending or descending node'''

    def __init__(self, inclination):
        super(SetInclinationTo, self).__init__()
        self.inclination = inclination

    def __apply__(self, orbit):
        orbit.i = self.inclination

    def __plot__(self, orbit, plotter, next_operation=None):
        self.__apply__(orbit)
        plotter._plot_apsides(orbit , label="Changed incalination")
    
    def velocity(self, orbit):
        with saved_state(orbit):
            orbit.f = 2 * pi - orbit.arg_pe 
            old_v = orbit.v
            
            self.__apply__(orbit)
            new_v = orbit.v

        return new_v - old_v
    
    def __repr__(self, r):
        r.positional_from_attr('inclination')

#########################################################################
#########################################################################

class ChangeInclanationBy(ReprHelperMixin, ImpulseOperation):
    '''Operation for changing inclination, when applied must be at ascending or descending node'''

    def __init__(self, delta):
        super(ChangeInclanationBy, self).__init__()
        self.delta = delta

    def __apply__(self, orbit):
        orbit.i += self.delta

    def __plot__(self, orbit, plotter, next_operation=None):
        self.__apply__(orbit)
        plotter._plot_apsides(orbit , label="Changed incalination")
    
    def velocity(self, orbit):
        with saved_state(orbit):
            orbit.f = 2 * pi - orbit.arg_pe 
            old_v = orbit.v
            
            self.__apply__(orbit)
            new_v = orbit.v

        return new_v - old_v
    
    def __repr__(self, r):
        r.positional_from_attr('delta')

#########################################################################
#ANOMALY
#########################################################################

class PropagateAnomalyTo(ReprHelperMixin, TimeOperation):
    def __init__(self, **kwargs):
        if len(kwargs) != 1:
            raise ValueError("Only one anomaly value must be passed.")

        key, value = list(kwargs.items())[0]
        if key not in ['M', 'E', 'f']:
            raise ValueError("Valid arguments are: M, E, f")

        self.key = key
        self.anomaly = value
        super().__init__(self.time_from_anomaly)

    def time_from_anomaly(self, orbit):
        if self.key == 'M':
            M = self.anomaly
        elif self.key == 'E':
            M = mean_anomaly_from_eccentric(orbit.e, self.anomaly)
        elif self.key == 'f':
            M = mean_anomaly_from_true(orbit.e, self.anomaly)

        if M < orbit.M:
            M += 2 * pi

        return (M - orbit.M) / orbit.n

    def __plot__(self, orbit, plotter, next_operation=None):
        f1 = orbit.f1
        orbit.t += self.time_delta(orbit)
        f2 = orbit.f

        plotter._plot_apsides(orbit , f2=f2 , propageted =True )

    def __repr__(self, r):
        r.keyword_with_value(self.key. self.anomaly)

#########################################################################
#########################################################################

class PropagateAnomalyBy(ReprHelperMixin, TimeOperation):
    '''Propagate to by the given  value'''

    def __init__(self, **kwargs):
        super(PropagateAnomalyBy, self).__init__()

        valid_args = ['M', 'E', 'f']
        extra_args = set(kwargs.keys()) - valid_args

        # Check if any extra arguments are provided
        if extra_args:
            raise ValueError(f"Invalid arguments: {extra_args}. Valid arguments are: {valid_args}")
        
        # Check if there is a value at all 
        if not kwargs:
            raise ValueError("At least one anomaly value must be passed.")
        
        # Check if only 1 value is passed
        if sum([1 for anomaly in kwargs.values() if anomaly is not None]) != 1:
            raise ValueError("Only one anomaly value must be passed.")
        

    def time(self, orbit):
        if self.key == 'f':
            orbits, f = divmod(self.anomaly, 2 * pi)
            M = mean_anomaly_from_true(orbit.e, f)
            return orbits * orbit.T + M / orbit.n
        elif self.key == 'E':
            orbits, E = divmod(self.anomaly, 2 * pi)
            M = mean_anomaly_from_eccentric(orbit.e, E)
            return orbits * orbit.T + M / orbit.n
        elif self.key == 'M':
            return self.anomaly / orbit.n
        
    def __plot__(self, orbit, plotter, next_operation=None):
        f1 = orbit.f
        orbit.t += self.time_delta(orbit)
        f2 = orbit.f

        plotter._plot_apsides(orbit , f2=f2 , propageted = True )

    def __repr__(self, r):
        r.keyword_with_value(self.key, self.anomaly)

#########################################################################
#ETC
#########################################################################

class Circularise(ReprHelperMixin, ImpulseOperation):
    '''Circularise the orbit, must be at apsides'''

    def __init__(self, raise_pericenter=True):
        super(Circularise, self).__init__()
        self.raise_pericenter = raise_pericenter

    def __apply__(self, orbit):
        if self.raise_pericenter:
            radius = orbit.apocenter_radius
        else:
            radius = orbit.pericenter_radius

        a, e = elements_4_apsides(radius, radius)

        orbit.a = a
        orbit.e = e

        orbit.v = orbit.v

    def __plot__(self, orbit, plotter, next_operation=None):
        self.__apply__(orbit)
        plotter._plot_apsides(orbit , label="Circularised orbit")
    
    def velocity(self, orbit):
        with saved_state(orbit):
          if self.raise_pericenter:
            orbit.propagate_anomaly_to(M=pi)
            radius = orbit.apocenter_radius
          else:
            orbit.propagate_anomaly_to(M=0)
            radius = orbit.pericenter_radius

        old_v = orbit.v

        a, e = elements_4_apsides(radius, radius)
        orbit.a = a
        orbit.e = e
        new_v = orbit.v

        return new_v - old_v
    
    def __repr__(self, r):
        r.keyword_with_attr('raise_pericenter')

#########################################################################
#########################################################################

class SetPerincenterHere(ReprHelperMixin, ImpulseOperation):
    '''Set pericenter here to the exact location, prep for eliptical oribt, must be from circular orbit'''

    def __init__(self, raise_pericenter=True):
        super(SetPerincenterHere, self).__init__()

    def __apply__(self, orbit):
        orbit.arg_pe = orbit.f
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            orbit.f = 0.0
    
    def __plot__(self, orbit, plotter, next_operation=None):
        self.__apply__(orbit)
        plotter._plot_apsides(orbit , label="Set pericenter here")
    
    def __repr__(self, r):
        pass


#########################################################################
#MANEUVER (Manover? Maneuver? ManRollover with a french accent?)
#########################################################################

class Maneuver(ReprHelperMixin, object):
    ''''Collection of operations '''

    def __init__(self, *operations):
        if not isinstance(operations, (list)):
            operations = [operations]

        self.operations = operations
    
#APOCENTER MANROLLOVERS

    @classmethod
    def set_apocenter_radius_to(cls, apocenter_radius):
        """Maneuver to set the apocenter radius."""
        operations = [
            PropagateAnomalyTo(M=pi),
            SetApocenterRadiusTo(apocenter_radius),
        ]
        return cls(operations)
    
    @classmethod
    def set_apocenter_altitude_to(cls, apocenter_altitude):
        """Maneuver to set the apocenter altitude."""
        operations = [
            PropagateAnomalyTo(M=pi),
            SetApocenterAltitudeTo(apocenter_altitude),
        ]
        return cls(operations)
    
    @classmethod
    def change_apocenter_by(cls, delta):
        """Maneuver to change the apocenter radius."""
        operations = [
            PropagateAnomalyTo(M=pi),
            ChangeApocenterBy(delta),
        ]
        return cls(operations)
    
#PERICENTER MANROLLOVERS

    @classmethod
    def set_pericenter_radius_to(cls, pericenter_radius):
        """Maneuver to set the pericenter radius."""
        operations = [
            PropagateAnomalyTo(M=0),
            SetPericenterRadiusTo(pericenter_radius),
        ]
        return cls(operations)
    
    @classmethod
    def set_pericenter_altitude_to(cls, pericenter_altitude):
        """Maneuver to set the pericenter altitude."""
        operations = [
            PropagateAnomalyTo(M=0),
            SetPericenterAltitudeTo(pericenter_altitude),
        ]
        return cls(operations)
    
    @classmethod
    def change_pericenter_by(cls, delta):
        """Maneuver to change the pericenter radius."""
        operations = [
            PropagateAnomalyTo(M=0),
            ChangePericenterBy(delta),
        ]
        return cls(operations)
    
#INCLINATION MANROLLOVERS

    @classmethod
    def set_inclination_to(cls, inclination):
        """Maneuver to set the inclination."""
        operations = [
            lambda orbit:
            PropagateAnomalyTo(f=2 *pi - orbit.arg_pe),
            SetInclinationTo(inclination),
        ]
        return cls(operations)

    @classmethod
    def change_inclination_by(cls, delta):
        """Maneuver to change the inclination."""
        operations = [
            lambda orbit:
            PropagateAnomalyTo(f=2 *pi - orbit.arg_pe),
            ChangeInclanationBy(delta),
        ]
        return cls(operations)
    
#HOHHMANN MANROLLOVERS
    
    @classmethod
    def hohmann_to_radius(cls, radius):
        """Maneuver to perform a Hohmann transfer to change the apocenter from pos."""
        operations = [
            SetPerincenterHere(),
            SetApocenterRadiusTo(radius),
            PropagateAnomalyTo(M=pi),
            Circularise(),
        ]
        return cls(operations)
    
    @classmethod
    def hohmann_to_altitude(cls, altitude):
        """Maneuver to perform a Hohmann transfer to change the apocenter from pos."""
        operations = [
            SetPerincenterHere(),
            SetApocenterAltitudeTo(altitude),
            PropagateAnomalyTo(M=pi),
            Circularise(),
        ]
        return cls(operations)
    


    def __apply__(self, orbit):
        for operation in self.operations:
            if callable(operation):
                operation = operation(orbit)
            if hasattr(operation, '__apply__') and callable(getattr(operation, '__apply__')):
                operation.__apply__(orbit)
            elif isinstance(operation, ImpulseOperation):
                orbit.v += operation.velocity(orbit)
            elif isinstance(operation, TimeOperation):
                orbit.t += operation.time(orbit)

    def __iapply__(self,orbit, copy=False):
        for operation in self.operations:
            if callable(operation):
                operation = operation(orbit)
            
            if copy:
                yield _copy(orbit), operation
            else:
                yield orbit, operation
            
            if hasattr(operation, '__apply__') and callable(getattr(operation, '__apply__')):
                operation.__apply__(orbit)
            elif isinstance(operation, ImpulseOperation):
                orbit.v += operation.velocity(orbit)
            elif isinstance(operation, TimeOperation):
                orbit.t += operation.time(orbit)

    def __repr__(self, r):
        r.positional_from_attr('operations')

    def __add__(self, other):
        """Add two maneuvers together."""
        if isinstance(other, Maneuver):
            return Maneuver(self.operations + other.operations)
        else:
            return NotImplementedError()



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.animation as animation
from matplotlib.widgets import Slider, Button, RadioButtons
import matplotlib
import inspect
from .objects import System
from . import objects
#from objects import System
#import objects
from collections import deque

matplotlib.use("TkAgg")

available_systems = objects.System.get_all_instances()
system_names = [sys.name for sys in available_systems]

# Creating the universe containing the desired system(s)
#sim = objects.Universe([objects.Solar_system])
#selected_system = objects.Solar_system


# Global State Variables (modified by UI or animation loop)
user_interacting = False
pan_info = {'button': None, 'start_x': None, 'start_y': None, 'current_xlim': None, 'current_ylim': None}
PAUSE = False
CURRENT_FRAME = 0
TRAIL_VISIBLE = True
NAMES_VISIBLE = False
c_scale = 1
scale = 1

# Global Plot Object Placeholders
fig, ax = None, None
plot_title = None
ani = None
sim = None
selected_system = None
bodies_to_plot = None
center_obj = None
c_display_radius = 0.0

plot_elements = []
trail_lines = []
trajectory_history = {}
plot_positions_this_frame = []

# UI Element Placeholders
radio_systems = None
slider_c_scale = None
slider_scale = None
button_trail = None
button_names = None

# Sticky Axes Variables
global_min_x, global_max_x = np.inf, -np.inf
global_min_y, global_max_y = np.inf, -np.inf
limits_initialized = False
padding = 0.15

# Constants
SECONDS_PER_DAY = 24 * 3600
START_DAY = 0
DEFAULT_END_DAY = 365*300 # Example default
DEFAULT_DAYS_PER_FRAME = 1 # Example default
INTERVAL_MS = 20
MAX_TRAIL_POINTS = None # Or your preferred default
TRAIL_COLOR = 'white'
TRAIL_LINEWIDTH = 0.5
TRAIL_ALPHA = 0.7

# Callback function for changing systems
def switch_system(label):
    global sim, bodies_to_plot, plot_elements, trail_lines, trajectory_history, center_obj, selected_system, CURRENT_FRAME

    # Reset the frame counter
    CURRENT_FRAME = 0

    available_systems = objects.System.get_all_instances()
    system_names = [sys.name for sys in available_systems]

    # Find the selected system by name
    selected_system = available_systems[system_names.index(label)]
    sim = objects.Universe([selected_system])  # Update the simulation with the new system

    # Clear all existing name texts from the axes
    for text in ax.texts:
        try:
            text.remove()
        except:
            pass

    # Reset plot elements and trajectory history
    for patch in plot_elements:
        try:
            # Remove any name text attached to the patch
            if hasattr(patch, 'name_text'):
                try:
                    patch.name_text.remove()
                except:
                    pass
            patch.remove()
        except ValueError:
            pass
    plot_elements.clear()

    for line in trail_lines:
        try:
            line.remove()
        except ValueError:
            pass
    trail_lines.clear()

    trajectory_history.clear()

    # Update bodies to plot based on the new system
    bodies_to_plot = sim.get_bodies()

    # Update center object for the new system
    if sim.all_elements and isinstance(sim.all_elements[0], objects.System):
        center_obj_candidate = sim.all_elements[0].center
        center_obj = next((b for b in bodies_to_plot if b.name == center_obj_candidate.name), None)
    else:
        center_obj = None

    # Reinitialize the simulation
    init()  # Call the init function to set up the new system

    # Restart the animation from frame 0
    ani.frame_seq = ani.new_frame_seq()
    print(f"Switched to system: {label}")

# Callback functions for sliders
def update_c_scale(val):
    global c_scale
    c_scale = val
    slider_c_scale.label.set_text(f'Center Body Scale  ({c_scale:.2f}): ')  # Update label text

    # Clear trail lines and history
    for line in trail_lines:
        line.set_data([], [])  # Clear the line data
    for i in range(len(trajectory_history)):
        trajectory_history[i].clear()  # Clear the history for each body

def update_scale(val):
    global scale
    scale = val
    slider_scale.label.set_text(f'Orbiting Bodies Scale ({scale:.2f}): ')  # Update label text


# Callback functions for buttons
def toggle_trail(event):
    global TRAIL_VISIBLE
    TRAIL_VISIBLE = not TRAIL_VISIBLE

def toggle_names(event):
    global NAMES_VISIBLE
    NAMES_VISIBLE = not NAMES_VISIBLE

def apply_global_limits_with_padding():
    """
    Applies the current global limits (with padding) to the axes of the plot.
    This ensures that the plotted area accommodates all bodies within the system,
    including a small buffer zone around the plot limits for better visualization.

    The function adjusts the x and y limits of the plot based on the minimum
    and maximum positions observed so far, applying a padding factor to prevent
    excessive zooming or clipping.
    """
    # Applies the current global limits (with padding) to the axes.
    global global_min_x, global_max_x, global_min_y, global_max_y
    if not limits_initialized:
        return # Don't apply if not initialized

    # Calculate range based on the *global* min/max data points seen so far
    x_range = global_max_x - global_min_x
    y_range = global_max_y - global_min_y

    # Add a small absolute value or fraction of the coordinate if non-zero
    min_abs_range = 1e9 # Minimum range to prevent excessive zoom on single points
    if x_range < 1e-9:
        x_range = max(abs(global_min_x) * 0.2, min_abs_range) if abs(global_min_x) > 1e-9 else min_abs_range
    if y_range < 1e-9:
        y_range = max(abs(global_min_y) * 0.2, min_abs_range) if abs(global_min_y) > 1e-9 else min_abs_range

    # Calculate desired padded limits based on global range
    target_xmin = global_min_x - x_range * padding
    target_xmax = global_max_x + x_range * padding
    target_ymin = global_min_y - y_range * padding
    target_ymax = global_max_y + y_range * padding

    # Set the limits on the axes using the calculated global targets
    ax.set_xlim(target_xmin, target_xmax)
    ax.set_ylim(target_ymin, target_ymax)

# Event Handlers

def on_scroll(event):
    """Handle mouse scroll for zooming."""
    global user_interacting
    if event.inaxes != ax: return

    user_interacting = True # User has taken control
    base_scale = 1.1 # Zoom factor
    cur_xlim = ax.get_xlim()
    cur_ylim = ax.get_ylim()
    xdata = event.xdata # Mouse x position
    ydata = event.ydata # Mouse y position

    if event.button == 'up': # Zoom in
        scale_factor = 1 / base_scale
    elif event.button == 'down': # Zoom out
        scale_factor = base_scale
    else: # Unexpected event
        scale_factor = 1

    new_width = (cur_xlim[1] - cur_xlim[0]) * scale_factor
    new_height = (cur_ylim[1] - cur_ylim[0]) * scale_factor

    relx = (cur_xlim[1] - xdata) / (cur_xlim[1] - cur_xlim[0])
    rely = (cur_ylim[1] - ydata) / (cur_ylim[1] - cur_ylim[0])

    ax.set_xlim([xdata - new_width * (1 - relx), xdata + new_width * relx])
    ax.set_ylim([ydata - new_height * (1 - rely), ydata + new_height * rely])
    ax.figure.canvas.draw_idle() # Request redraw

def on_press(event):
    """Handle mouse button press for panning."""
    global user_interacting, pan_info
    if event.inaxes != ax: return
    # Use middle mouse button for panning (button 2), or left (button 1) if preferred
    if event.button != 1: return

    user_interacting = True # User has taken control
    pan_info['button'] = event.button
    pan_info['start_x'] = event.xdata
    pan_info['start_y'] = event.ydata
    pan_info['current_xlim'] = ax.get_xlim()
    pan_info['current_ylim'] = ax.get_ylim()

def on_motion(event):
    """Handle mouse motion while button is pressed for panning."""
    global pan_info
    if event.inaxes != ax or pan_info['button'] != event.button: return
    if pan_info['start_x'] is None or event.xdata is None: return # Avoid errors if mouse leaves axes

    dx = event.xdata - pan_info['start_x']
    dy = event.ydata - pan_info['start_y']

    new_xlim = (pan_info['current_xlim'][0] - dx, pan_info['current_xlim'][1] - dx)
    new_ylim = (pan_info['current_ylim'][0] - dy, pan_info['current_ylim'][1] - dy)

    ax.set_xlim(new_xlim)
    ax.set_ylim(new_ylim)
    ax.figure.canvas.draw_idle() # Request redraw

def on_release(event):
    """Handle mouse button release to stop panning."""
    global pan_info
    if event.button == pan_info['button']:
        pan_info = {'button': None, 'start_x': None, 'start_y': None, 'current_xlim': None, 'current_ylim': None}

def on_key(event):
    """Handle key press events, e.g., reset view."""
    global user_interacting, global_min_x, global_max_x, global_min_y, global_max_y, PAUSE, c_scale
    if event.key == 'r': # 'r' for reset
        user_interacting = False
        # Recalculate limits based on *current* positions to avoid jumping
        # This requires getting current positions, similar to animate()
        frame_min_x, frame_max_x = np.inf, -np.inf
        frame_min_y, frame_max_y = np.inf, -np.inf

        # Need current center position if shifting is enabled
        c_x_now, c_y_now = (0.0, 0.0)
        c_display_radius = (center_obj.mean_diameter / 2.0) * c_scale

        if center_obj:
            c_x_now = center_obj.x
            c_y_now = center_obj.y

        for i, body in enumerate(bodies_to_plot):
             x_orig = body.x
             y_orig = body.y
             # Use stored radius if possible, otherwise recalculate
             try:
                 display_radius = plot_elements[i].get_radius()
             except (IndexError, AttributeError):
                 display_radius = (body.mean_diameter / 2.0) * scale # Fallback

             plot_x, plot_y = x_orig, y_orig
             if center_obj and body.name != center_obj.name:
                 dx = x_orig - c_x_now
                 dy = y_orig - c_y_now
                 distance_from_center = np.sqrt(dx**2 + dy**2)
                 if distance_from_center > 1e-9 and c_display_radius > 1e-9:
                     ux = dx / distance_from_center
                     uy = dy / distance_from_center
                     plot_x = x_orig + ux * c_display_radius
                     plot_y = y_orig + uy * c_display_radius

             frame_min_x = min(frame_min_x, plot_x - display_radius)
             frame_max_x = max(frame_max_x, plot_x + display_radius)
             frame_min_y = min(frame_min_y, plot_y - display_radius)
             frame_max_y = max(frame_max_y, plot_y + display_radius)

        if not np.isinf(frame_min_x):
            global_min_x, global_max_x = frame_min_x, frame_max_x
            global_min_y, global_max_y = frame_min_y, frame_max_y
            apply_global_limits_with_padding() # Apply the reset limits
            ax.figure.canvas.draw_idle() # Request redraw
        else:
            print("Warning: Could not determine current limits for reset.")

    if event.key == ' ':
        PAUSE = not PAUSE
        if PAUSE:
            ani.event_source.stop()
        else:
            ani.event_source.start()
def init():
    """
    Initializes the plot and simulation state at the start of the animation.
    This function prepares the plot by creating the bodies and setting up their
    initial positions and sizes. It also sets the initial axes limits and prepares
    the system for the first frame of the animation.

    Returns:
        list: A list of plot elements (circles, trails, title) that are created for the
              first frame of the simulation.
    """
    global global_min_x, global_max_x, global_min_y, global_max_y, limits_initialized, user_interacting, c_scale, bodies_to_plot

    user_interacting = False

    for text in ax.texts:
        try:
            text.remove()
        except:
            pass

    # Reset plot elements and trajectory history
    for patch in plot_elements:
        try:
            # Remove any name text attached to the patch
            if hasattr(patch, 'name_text'):
                try:
                    patch.name_text.remove()
                except:
                    pass
            patch.remove()
        except ValueError:
            pass
    plot_elements.clear()

    for line in trail_lines:
        try:
            line.remove()
        except ValueError:
            pass
    trail_lines.clear()

    trajectory_history.clear()

    bodies_to_plot = sim.get_bodies()

    # --- Run simulation for the very first frame (t=0) ---
    sim.update(START_DAY * SECONDS_PER_DAY, logarithmic=False)

    # --- Get initial positions and limits for t=0 ---
    min_x_init, max_x_init = np.inf, -np.inf
    min_y_init, max_y_init = np.inf, -np.inf

    # Get center's position at t=0
    c_x_init, c_y_init = (0.0, 0.0)
    if center_obj:
        c_x_init = center_obj.x
        c_y_init = center_obj.y
        c_display_radius = (center_obj.mean_diameter / 2.0) * c_scale

    # --- Create circle patches for t=0 and calculate initial limits ---
    for i, body in enumerate(bodies_to_plot):
        x_orig = body.x # Position at t=0
        y_orig = body.y # Position at t=0
        color = body.color

        # Determine radius and initial plot position (with shift at t=0)
        if center_obj and body.name == center_obj.name:
            display_radius = c_display_radius
            z_order = 10
            plot_x, plot_y = x_orig, y_orig
        else:
            display_radius = (body.mean_diameter / 2.0) * scale
            z_order = 20
            # Shift calculation relative to center's t=0 position
            dx = x_orig - c_x_init
            dy = y_orig - c_y_init
            distance_from_center = np.sqrt(dx ** 2 + dy ** 2)
            if distance_from_center > 1e-9:
                ux = dx / distance_from_center
                uy = dy / distance_from_center
                plot_x = x_orig + ux * c_display_radius
                plot_y = y_orig + uy * c_display_radius
            else:
                plot_x, plot_y = x_orig, y_orig

        # Update limits based on initial positions/sizes
        min_x_init = min(min_x_init, plot_x - display_radius)
        max_x_init = max(max_x_init, plot_x + display_radius)
        min_y_init = min(min_y_init, plot_y - display_radius)
        max_y_init = max(max_y_init, plot_y + display_radius)

        # Create the circle at t=0 POSITION and add it
        circle = patches.Circle((plot_x, plot_y), radius=display_radius, color=color, zorder=z_order)
        ax.add_patch(circle)
        plot_elements.append(circle) # Store the circle object

        # Initialize Trail
        trajectory_history[i] = deque(maxlen=MAX_TRAIL_POINTS)
        trajectory_history[i].append((plot_x, plot_y))
        line, = ax.plot([], [], color=TRAIL_COLOR, linewidth=TRAIL_LINEWIDTH, alpha=TRAIL_ALPHA,
                        zorder=z_order - 1)  # Trail behind body
        trail_lines.append(line)  # Store the line object

    # Set initial axes limits for t=0
    if np.isinf(min_x_init): min_x_init, max_x_init, min_y_init, max_y_init = -1e10, 1e10, -1e10, 1e10 # Fallback

    global_min_x, global_max_x = min_x_init, max_x_init
    global_min_y, global_max_y = min_y_init, max_y_init
    limits_initialized = True

    apply_global_limits_with_padding()

    # Set initial title
    plot_title.set_text(f"{selected_system.name} system at t = {START_DAY:.1f} days ({START_DAY/365.2425:.2f} years)")
    return plot_elements + trail_lines + [plot_title]


def animate(frame):
    """
    Updates the plot for each frame of the animation. The function recalculates the
    positions of the bodies in the system based on the current simulation time and updates
    the plot elements (bodies and trails) accordingly. The axes limits are also adjusted
    dynamically based on the bodies' positions to ensure they stay within view.

    Args:
        frame (int): The current frame number of the animation. This determines the
                     current simulation time and updates the plot accordingly.

    Returns:
        list: A list of plot elements (circles, trails, title) that are updated for the
              current frame of the simulation.
    """
    global CURRENT_FRAME, global_min_x, global_max_x, global_min_y, global_max_y, plot_positions_this_frame, scale, c_scale, NAMES_VISIBLE, DAYS_PER_FRAME

    # Update the frame counter
    CURRENT_FRAME = frame
    # Calculate current simulation time based on frame number and step
    current_day = START_DAY + frame * DAYS_PER_FRAME
    current_time_seconds = current_day * SECONDS_PER_DAY

    # Update simulation state
    sim.update(current_time_seconds, logarithmic=False)

    # Get center's current position for shifting calculations in this frame
    c_x_now, c_y_now = (0.0,0.0)
    if center_obj:
        c_x_now = center_obj.x
        c_y_now = center_obj.y
        c_display_radius = (center_obj.mean_diameter / 2.0) * c_scale

    # Clear previous frame's positions
    plot_positions_this_frame.clear()

    # Get the updates positions and calculate new axes limits
    for i, body in enumerate(bodies_to_plot):
        if i >= len(plot_elements):  # Safety check
            print(
                f"Warning: Mismatch between bodies_to_plot ({len(bodies_to_plot)}) and plot_elements ({len(plot_elements)}) at index {i}")
            continue
        circle = plot_elements[i]  # Get the corresponding circle patch
        line = trail_lines[i]

        x_orig = body.x
        y_orig = body.y

        if body.name == center_obj.name:
            display_radius = (body.mean_diameter / 2.0) * c_scale  # Center body
        else:
            display_radius = (body.mean_diameter / 2.0) * scale  # Orbiting bodies

        plot_x, plot_y = x_orig, y_orig
        if center_obj and body.name != center_obj.name:
            dx = x_orig - c_x_now
            dy = y_orig - c_y_now
            distance_from_center = np.sqrt(dx**2 + dy**2)
            if distance_from_center > 1e-9 and c_display_radius > 1e-9:
                ux = dx / distance_from_center
                uy = dy / distance_from_center
                plot_x = x_orig + ux * c_display_radius
                plot_y = y_orig + uy * c_display_radius

        # Update the circle's center position
        circle.set_center((plot_x, plot_y))
        circle.set_radius(display_radius)
        plot_positions_this_frame.append((plot_x, plot_y, display_radius)) # Store for limits/reset

        # --- Update Trail ---
        # Append current *plotted* position to this body's history
        trajectory_history[i].append((plot_x, plot_y))
        # Extract history for plotting (handle case of < 2 points)
        hist = list(trajectory_history[i])  # Convert deque to list for slicing/zipping
        if len(hist) >= 2:
            hist_x, hist_y = zip(*hist)
            line.set_data(hist_x, hist_y)
        elif len(hist) == 1:  # Only one point
            line.set_data([hist[0][0]], [hist[0][1]])  # Plot single point
        else:  # No history yet
            line.set_data([], [])

        # Update trail visibility
        if TRAIL_VISIBLE:
            line.set_visible(True)
        else:
            line.set_visible(False)

        # Update Names
        if NAMES_VISIBLE:
            # Check if the body is within axes limits
            if ax.get_xlim()[0] <= plot_x <= ax.get_xlim()[1] and ax.get_ylim()[0] <= plot_y <= ax.get_ylim()[1]:
                # Create or update the text label for the body
                if not hasattr(circle, 'name_text'):
                    # Create a new text object if it doesn't exist
                    circle.name_text = ax.text(plot_x, plot_y, body.name, color='white', fontsize=8, ha='center',
                                               va='bottom', zorder=30)
                else:
                    # Update the position of the existing text object
                    circle.name_text.set_position((plot_x, plot_y + display_radius))
                    circle.name_text.set_text(body.name)  # Update the name text
            else:
                # If names are not visible, remove any existing text
                if hasattr(circle, 'name_text'):
                    circle.name_text.remove()
                    del circle.name_text
        else:
            # If names are not visible, remove any existing text
            if hasattr(circle, 'name_text'):
                circle.name_text.remove()
                del circle.name_text

        # Handle Axes Limits
    if not user_interacting:
            # Calculate frame limits ONLY if not user interacting
        frame_min_x, frame_max_x = np.inf, -np.inf
        frame_min_y, frame_max_y = np.inf, -np.inf
        for px, py, radius in plot_positions_this_frame:
            frame_min_x = min(frame_min_x, px - radius)
            frame_max_x = max(frame_max_x, px + radius)
            frame_min_y = min(frame_min_y, py - radius)
            frame_max_y = max(frame_max_y, py + radius)

        if not np.isinf(frame_min_x):  # Ensure we got valid numbers this frame
            global_min_x = min(global_min_x, frame_min_x)
            global_max_x = max(global_max_x, frame_max_x)
            global_min_y = min(global_min_y, frame_min_y)
            global_max_y = max(global_max_y, frame_max_y)

        # Apply the potentially updated global limits with padding
        # This ensures the axes only expand or stay the same, never shrink
        apply_global_limits_with_padding()

    # Update the title
    plot_title.set_text(f"{selected_system.name} at t = {current_day:.1f} days ({current_day/365.2425:.2f} years)")

        # Return iterable of plot elements that have been updated
    return plot_elements + trail_lines + [plot_title]

# List available systems
def list_systems():
    #print("Available systems:", system_names)
    available_systems = objects.System.get_all_instances()
    system_names = [sys.name for sys in available_systems]
    return system_names

# Plot a system (from outside)
def plot(system, END_DAY = 365, switching = True):
    run_animation_gui(system, END_DAY, DEFAULT_DAYS_PER_FRAME, switching)


# --- Main function to set up and run the animation GUI ---
def run_animation_gui(default_system_name="Solar system", END_DAY=DEFAULT_END_DAY,
                      days_per_frame=DEFAULT_DAYS_PER_FRAME, switching = True):
    global fig, ax, plot_title, ani, sim, selected_system, bodies_to_plot, center_obj, c_display_radius
    global radio_systems, slider_c_scale, slider_scale, button_trail, button_names
    global c_scale, scale, limits_initialized, CURRENT_FRAME, PAUSE, user_interacting
    global plot_elements, trail_lines, trajectory_history, plot_positions_this_frame
    global global_min_x, global_max_x, global_min_y, global_max_y  # For resetting limits

    # Reset states for a fresh run or system switch
    CURRENT_FRAME = 0
    PAUSE = False
    user_interacting = False
    limits_initialized = False
    global_min_x, global_max_x = np.inf, -np.inf
    global_min_y, global_max_y = np.inf, -np.inf

    # Clear plot element lists that might persist from a previous run (if any)
    plot_elements.clear()
    trail_lines.clear()
    trajectory_history.clear()
    plot_positions_this_frame.clear()

    # Set initial scales (these are module globals, reset them for this run)
    # These could also be arguments to run_animation_gui if you want more control
    initial_c_scale_val = 1
    initial_scale_val = 1
    c_scale = initial_c_scale_val  # Update global c_scale
    scale = initial_scale_val  # Update global scale

    available_systems = objects.System.get_all_instances()
    system_names = [sys.name for sys in available_systems]

    # Find the selected system
    try:
        if not available_systems:  # Should have been populated on import
            print("Error: No systems available from objects.py.")
            return
        selected_system_obj = available_systems[system_names.index(default_system_name)]
    except (ValueError, IndexError):
        print(f"Warning: Default system '{default_system_name}' not found. Using first available: {system_names[0]}")
        selected_system_obj = available_systems[0]

    selected_system = selected_system_obj  # Assign to global
    sim = objects.Universe([selected_system_obj])  # Assign to global
    bodies_to_plot = sim.get_bodies()  # Assign to global

    # Plotting setup
    fig, ax = plt.subplots(figsize=(15, 15))  # Assign to global fig, ax
    fig.set_facecolor('black')
    ax.set_facecolor('black')
    ax.set_aspect('equal', adjustable='box')
    ax.set_xlabel("X (m)", color="white")
    ax.set_ylabel("Y (m)", color="white")
    plot_title = ax.set_title("", color="white")  # Assign to global
    ax.tick_params(axis='x', colors='white')
    ax.tick_params(axis='y', colors='white')
    ax.spines['bottom'].set_color('white')
    ax.spines['top'].set_color('white')
    ax.spines['left'].set_color('white')
    ax.spines['right'].set_color('white')

    # Heuristic to find the "main" center
    center_obj = None  # Reset global center_obj before finding new one
    if sim.all_elements and isinstance(sim.all_elements[0], objects.System):
        center_obj_candidate = sim.all_elements[0].center
        center_obj = next((b for b in bodies_to_plot if b.name == center_obj_candidate.name), None)

    if center_obj:
        # c_display_radius is calculated in init/animate based on global c_scale
        pass  # No need to set c_display_radius here, init/animate will use global c_scale
    else:
        print("Warning: Center body for shifting not definitively identified. Shifting might be inconsistent.")
        c_display_radius = 0.0  # Ensure it's zero if no center

    # Create UI elements using the new fig
    # Radio buttons
    # Use fig.add_axes for more control if fig is created here
    if switching:
        ax_systems_ui = fig.add_axes([0.8, 0.3, 0.15, 0.15])
        radio_systems = RadioButtons(ax_systems_ui, system_names,
                                 active=system_names.index(
                                     default_system_name) if default_system_name in system_names else 0)
        radio_systems.on_clicked(switch_system)  # switch_system is a global function

    # Sliders
    ax_c_scale_ui = fig.add_axes([0.8, 0.14, 0.15, 0.03])
    slider_c_scale = Slider(ax_c_scale_ui, ' ', 1, 5000.0, valinit=c_scale, color="white")
    slider_c_scale.label.set_color('white')
    slider_c_scale.label.set_text(f'Center Body Scale  ({c_scale:.2f}): ')
    slider_c_scale.on_changed(update_c_scale)  # update_c_scale is a global function

    ax_scale_ui = fig.add_axes([0.8, 0.1, 0.15, 0.03])
    slider_scale = Slider(ax_scale_ui, ' ', 1, 5000.0, valinit=scale, color="white")
    slider_scale.label.set_color('white')
    slider_scale.label.set_text(f'Orbiting Bodies Scale ({scale:.2f}): ')
    slider_scale.on_changed(update_scale)  # update_scale is a global function

    # Buttons
    ax_toggle_trail_ui = fig.add_axes([0.8, 0.199, 0.15, 0.04])
    button_trail = Button(ax_toggle_trail_ui, 'Trails Visibility')
    button_trail.on_clicked(toggle_trail)  # toggle_trail is a global function

    ax_toggle_names_ui = fig.add_axes([0.8, 0.249, 0.15, 0.04])
    button_names = Button(ax_toggle_names_ui, 'Names Visibility')
    button_names.on_clicked(toggle_names)  # toggle_names is a global function

    # Connect Event Handlers to the new fig
    fig.canvas.mpl_connect('scroll_event', on_scroll)
    fig.canvas.mpl_connect('button_press_event', on_press)
    fig.canvas.mpl_connect('motion_notify_event', on_motion)
    fig.canvas.mpl_connect('button_release_event', on_release)
    fig.canvas.mpl_connect('key_press_event', on_key)

    # Animation
    # Use END_DAY and days_per_frame passed as arguments
    num_frames = int((END_DAY - START_DAY) / days_per_frame) + 1
    # Update global DAYS_PER_FRAME if it's used directly in animate/init
    # It's better if animate/init get this info or it's a well-defined global
    # For now, animate uses global DAYS_PER_FRAME, so let's ensure it's set if changed by args
    global DAYS_PER_FRAME  # Make sure we are affecting the global one if animate uses it
    DAYS_PER_FRAME = days_per_frame  # Set the global for animate to use

    print(f"Starting animation: {num_frames} frames, {DAYS_PER_FRAME} day(s) per frame, interval={INTERVAL_MS}ms")
    print("\n--- Interactive Controls ---")
    print("Mouse Wheel: Zoom in/out")
    print("Left Mouse Button + Drag: Pan")
    print("Press 'space': Pause/Unpause animation")
    print("Press 'r': Reset view")
    print("--------------------------\n")

    ani = animation.FuncAnimation(fig, animate, frames=num_frames,
                                  init_func=init, blit=False, interval=INTERVAL_MS,
                                  repeat=True)  # Assign to global ani
    plt.show()  # This is blocking
    print("Animation finished or window closed.")

if __name__ == "__main__":
    # This block runs ONLY when the script is executed directly
    run_animation_gui()



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import PyQt5 as qt

'''
Most important piece of code, tests if testing is testing

Args:
    string (str): A string to test the testing function.
Returns:
    str: A string indicating that the testing function is working.
    

'''

def testingtesting(string):
    answ =  f"I'm testing this very much: {string}"
    return answ




# Using code from http://www.jgiesen.de/kepler/kepler.html
# Implemnetation based on the following paper: [A Practical Method for Solving the Kepler Equation][1] by Marc A. Murison from the U.S. Naval Observatory
# [1] https://www.researchgate.net/publication/271214938_A_Practical_Method_for_Solving_the_Kepler_Equation

#Compatibility reasons
from __future__ import absolute_import, division, print_function

from collections import namedtuple
from contextlib import contextmanager
from copy import deepcopy
from typing import Literal
from dataclasses import dataclass
import numpy as np

# from numpy import atan2, floor, fmod, ising, isnan
# from numpy import arrcos as acos
# from numpy import cos, dot, sin, sqrt
# from numpy.linalg import norm
# #from math import atan2, floor, fmod, ising, isnan #chatqpt advice
#from scipy.constants import pi

MAX_ITER = 100

# Custom exception Exception

class ConvergenceError(Exception):
    pass


#Anomaly class
'''
class Anomaly(object):
    '''
   # Class representing an anomaly (mean, eccentric, or true)
'''
    def __init__(self, **kwargs):
        super()().__init__()
        
        valid_args = set(['M' , 'E' , 'theta'])

        extra_args = set(kwargs.keys()) - valid_args
        if extra_args:
            raise TypeError('Invalid arguments: {}'.format(extra_args))
        if not kwargs:
            raise TypeError('No arguments provided.')
        
    def M(self, e):
        if self.key == 'M':
            return self.anomaly
        if self.key == 'E':
            return mean_anomaly_from_eccentric(e, self.anomaly)
        if self.key == 'theta':
            return mean_anomaly_from_true(e, self.anomaly)
        
    def E(self, e):
        if self.key == 'M':
            return eccentric_anomaly_from_mean(e, self.anomaly)
        elif self.key == 'E':
            return self.anomaly
        elif self.key == 'theta':
            return eccentric_anomaly_from_true(e, self.anomaly)
        
    def theta(self, e):
        if self.key == 'M':
            return true_anomaly_from_mean(e, self.anomaly)
        elif self.key == 'E':
            return true_anomaly_from_eccentric(e, self.anomaly)
        elif self.key == 'theta':
            return self.anomaly
'''

@dataclass
class Anomaly:
    """
    Represents an orbital anomaly and provides conversions between 
    mean (M), eccentric (E), and true (theta) anomalies.

    Attributes:
        value (float): The value of the anomaly.
        type (Literal['M', 'E', 'theta']): The type of the anomaly:
            - 'M': Mean anomaly
            - 'E': Eccentric anomaly
            - 'theta': True anomaly

    Methods:
        M(e: float) -> float:
            Returns the mean anomaly, computed from the stored value if necessary.

        E(e: float) -> float:
            Returns the eccentric anomaly, computed from the stored value if necessary.

        theta(e: float) -> float:
            Returns the true anomaly, computed from the stored value if necessary.
    """

    value: float
    type: Literal['M', 'E', 'theta']

    def __post_init__(self):
        valid_types = {'M', 'E', 'theta'}
        if self.type not in valid_types:
            raise ValueError(f"Invalid anomaly type: '{self.type}'. Must be one of {valid_types}.")

    def M(self, e: float) -> float:
        if self.type == 'M':
            return self.value
        elif self.type == 'E':
            return mean_anomaly_from_eccentric(e, self.value)
        elif self.type == 'theta':
            return mean_anomaly_from_true(e, self.value)

    def E(self, e: float) -> float:
        if self.type == 'M':
            return eccentric_anomaly_from_mean(e, self.value)
        elif self.type == 'E':
            return self.value
        elif self.type == 'theta':
            return eccentric_anomaly_from_true(e, self.value)

    def theta(self, e: float) -> float:
        if self.type == 'M':
            return true_anomaly_from_mean(e, self.value)
        elif self.type == 'E':
            return true_anomaly_from_eccentric(e, self.value)
        elif self.type == 'theta':
            return self.value


#Anomaly calcs

'''
def eccentric_anomaly_from_mean(e, M, tolerance=1e-10):

    Calc eccenetric anomaly from mean anomaly

    Mnorm = np.fmod(M, 2 * np.pi)
    E0 = M + (-1 / 2 * e**3 + e + (e**2 + 3 / 2 * np.cos(M) * e ** 3) * np.cos(M)) * np.sin(M)
    dE = tolerance + 1
    count = 0
    while dE > tolerance:
        t1 = np.cos(E0)
        t2 = -1 + e * t1
        t3 = np.sin(E0)
        t4 = e * t3
        t5 = -E0 + t4 + Mnorm
        t6 = t5 / (1 / 2 * t5 * t4 / t2 + t2)
        E = E0 - t5 / ((1 / 2 * t3 - 1 / 6 * t1 * t6) * e * t6 + t2)
        dE = abs(E - E0)
        E0 = E
        count += 1
        if count == MAX_ITER:
            raise ConvergenceError('Did not converge after {n} iterations. (e={e!r}, M={M!r})'.format(n=MAX_ITER, e=e, M=M))
        return E
'''

def eccentric_anomaly_from_mean(e, M, tolerance=1e-8):  # Using relaxed tolerance too
    """
    Compute the eccentric anomaly from the mean anomaly using the Newton-Raphson method.

    Args:
        e (float): Orbital eccentricity (0 <= e < 1).
        M (float): Mean anomaly in radians.
        tolerance (float, optional): Convergence tolerance. Default is 1e-8.

    Returns:
        float: Eccentric anomaly in radians.

    Raises:
        ConvergenceError: If the Newton-Raphson iteration fails to converge within MAX_ITER iterations.
    """
    Mnorm = np.fmod(M, 2 * np.pi)

    # Initial guess (Mnorm is usually good for Newton-Raphson)
    E0 = Mnorm
    # Optional refinement for high eccentricity, but Mnorm often suffices
    # if e > 0.8:
    # E0 = np.pi

    dE = tolerance + 1
    count = 0
    E = E0  # Start with E = E0


    while dE > tolerance:
        # Newton-Raphson step: E_new = E - f(E) / f'(E)
        # f(E) = E - e * sin(E) - Mnorm
        # f'(E) = 1 - e * cos(E)
        E_new = E - (E - e * np.sin(E) - Mnorm) / (1.0 - e * np.cos(E))

        dE = abs(E_new - E)
        E = E_new  # Update E for the next iteration
        count += 1

        if count == MAX_ITER:
            raise ConvergenceError(
                f'Newton-Raphson did not converge after {MAX_ITER} iterations. '
                f'(e={e!r}, M={Mnorm!r}, E={E!r}, dE={dE!r})'  # More info
            )
    return E

def eccentric_anomaly_from_true(e, theta):
    """
    Compute the eccentric anomaly from the true anomaly.

    Args:
        e (float): Orbital eccentricity (0 <= e < 1).
        theta (float): True anomaly in radians.

    Returns:
        float: Eccentric anomaly in radians.
    """
    E = np.atan2(np.sqrt(1 - e**2) * np.sin(theta), np.cos(theta) + e)
    E = np.mod(E, 2 * np.pi)
    return E

def mean_anomaly_from_eccentric(e, E):
    """
    Compute the mean anomaly from the eccentric anomaly.

    Args:
        e (float): Orbital eccentricity (0 <= e < 1).
        E (float): Eccentric anomaly in radians.

    Returns:
        float: Mean anomaly in radians.
    """
    M = E - e * np.sin(E)
    M = np.mod(M, 2 * np.pi)
    return M


def mean_anomaly_from_true(e, theta):
    """
    Compute the mean anomaly from the true anomaly.

    Args:
        e (float): Orbital eccentricity (0 <= e < 1).
        theta (float): True anomaly in radians.

    Returns:
        float: Mean anomaly in radians.
    """
    E = eccentric_anomaly_from_true(e, theta)
    M = E - e * np.sin(E)
    M = np.mod(M, 2 * np.pi)
    return M

def true_anomaly_from_eccentric(e, E):
    """
    Compute the true anomaly from the eccentric anomaly.

    Args:
        e (float): Orbital eccentricity (0 <= e < 1).
        E (float): Eccentric anomaly in radians.

    Returns:
        float: True anomaly in radians.
    """
    #theta = np.atan2(np.sqrt(1 - e**2) * np.sin(E), np.cos(E) + e)
    theta = np.arctan2(np.sqrt(1 - e ** 2) * np.sin(E), np.cos(E) + e) #Z
    theta = np.mod(theta, 2 * np.pi)
    return theta

def true_anomaly_from_mean(e, M, tolerance=1e-10):
    """
    Compute the true anomaly from the mean anomaly.

    This function internally computes the eccentric anomaly via Newton-Raphson, 
    then converts it to the true anomaly.

    Args:
        e (float): Orbital eccentricity (0 <= e < 1).
        M (float): Mean anomaly in radians.
        tolerance (float, optional): Convergence tolerance for Newton-Raphson. Default is 1e-10.

    Returns:
        float: True anomaly in radians.

    Raises:
        ConvergenceError: If Newton-Raphson fails to converge within MAX_ITER iterations.
    """
    E = eccentric_anomaly_from_mean(e, M, tolerance)
    theta = true_anomaly_from_eccentric(e, E)
    return theta

















"""
Orbital and Physical Constants for Solar System Bodies

This module defines key astronomical and planetary constants for the Sun, Moon,
and major Solar System objects, based on authoritative sources such as the IAU
2009 resolutions and NASA planetary fact sheets.

Constants include:
- Speed of light and universal gravitational constant
- Masses and gravitational parameters (μ) for the Sun, planets, and dwarf planets
- Equatorial, polar, and mean radii (in kilometers)
- Orbital eccentricities
- Perihelion distances (in kilometers)
- Orbital periods (in seconds)

Sources:
- Masses and gravitational constants: IAU 2009 recommendations  
https://iau-a3.gitlab.io/NSFA/IAU2009_consts.html
- Radii and orbital elements:  
https://orbital-mechanics.space/reference/planetary-parameters.html  
https://nssdc.gsfc.nasa.gov/planetary/factsheet/  
https://nssdc.gsfc.nasa.gov/planetary/factsheet/plutofact.html

All distances are in kilometers, times in seconds, and masses in kilograms unless otherwise noted.

Intended for use in orbital mechanics simulations, ephemeris generation, and astrodynamics analysis.

Example:
    from solar_system_constants import Earth_mu, Mars_e, Jupiter_T
"""

#IAU constants (source for mass: https://iau-a3.gitlab.io/NSFA/IAU2009_consts.html)
#Equitorial radii (sources: https://orbital-mechanics.space/reference/planetary-parameters.html, https://nssdc.gsfc.nasa.gov/planetary/factsheet/plutofact.html)

c = 299792458 #speed of light (m/s)

G = 6.67428e-11  #gravitational constant (m^3/kg*s^2)

S_M_P = 1.32712440041e20  #Solar mass parameter (m^3/s^2)
Sun_m = S_M_P / G #mass of Sun (kg)

Earth_mu = 3.986004415e14  #Geocentric gravitational constant (m^3/s^2)
Earth_m  = Earth_mu / G #mass of Earth (kg)
Earth_re = 6378.1366 #equatorial radius of Earth (km)
Earth_rp = 6356.8 #polar radius of Earth (km)
Earth_rm = 6371 #mean radius of Earth (km)
Earth_perihelion   = 147098290  #km perihelion distance from Sun
Earth_T   = 365.256 * 24 * 3600     # 31_558_118.4 seconds

Moon_rm = 1737.4           # mean radius (km)
Moon_m = 7.342e22       # mass in kilograms
Moon_e = 0.0549            # orbital eccentricity
Moon_perihelion = 363300 #km
Moon_T    = 27.3217 * 24 * 3600     # 2_360_591.68 seconds

#masses (kg)
Mercury_m = Sun_m / 6.02365733e6
Venus_m   = Sun_m / 4.08523719e5
Mars_m    = Sun_m / 3.09870359e6
Jupiter_m = Sun_m / 1.047348644e3
Saturn_m  = Sun_m / 3.4979018e3
Uranus_m  = Sun_m / 2.2902951e4
Neptune_m = Sun_m / 1.941226e4
Pluto_m   = Sun_m / 1.3605e8
Eris_m    = Sun_m / 1.191e8

#gravitational constants (m^3/s^2)
Mercury_mu = Mercury_m * G
Venus_mu   = Venus_m   * G
Mars_mu    = Mars_m    * G
Jupiter_mu = Jupiter_m * G
Saturn_mu  = Saturn_m  * G
Uranus_mu  = Uranus_m  * G
Neptune_mu = Neptune_m * G
Pluto_mu   = Pluto_m * G
Eris_mu    = Eris_m * G

#equatorial radii (km)
Sun_re = 695700
Mercury_re = 2440.53
Venus_re   = 6051.8
Mars_re    = 3396.19
Jupiter_re = 71492
Saturn_re  = 60268
Uranus_re  = 25559
Neptune_re = 24764
Pluto_re   = 1188.3
Eris_re    = 1163

#polar radii (km)
Sun_rp = Sun_re
Mercury_rp = 2438.26
Venus_rp   = Venus_re
Mars_rp    = 3376.2
Jupiter_rp = 66854
Saturn_rp  = 54364
Uranus_rp  = 24973
Neptune_rp = 24341
Pluto_rp   = Pluto_re
Eris_rp    = Eris_re

#mean radii (km)
Sun_rm = Sun_rp
Mercury_rm = 2439.4
Venus_rm   = Venus_rp
Mars_rm    = 3389.5
Jupiter_rm = 69911
Saturn_rm  = 58232
Uranus_rm  = 25362
Neptune_rm = 24622
Pluto_rm   = Pluto_rp
Eris_rm    = Eris_rp

#perihelion radius
Mercury_perihelion = 46001200     # km
Venus_perihelion   = 107476000    # km
Mars_perihelion    = 206700000    # km
Jupiter_perihelion = 740573600    # km
Saturn_perihelion  = 1353572956   # km
Uranus_perihelion  = 2734998229   # km
Neptune_perihelion = 4452940833   # km
Pluto_perihelion   = 4436750000   # km
Eris_perihelion    = 5765740000   # km

#Orbital eccentricity
Mercury_e = 0.2056
Venus_e   = 0.0068
Earth_e   = 0.0167
Mars_e    = 0.0934
Jupiter_e = 0.0489
Saturn_e  = 0.0565
Uranus_e  = 0.0457
Neptune_e = 0.0113
Pluto_e   = 0.2488
Eris_e    = 0.4418

#Periodic time (s)
Mercury_T = 87.97 * 24 * 3600       # 7_600_128 seconds
Venus_T   = 224.70 * 24 * 3600      # 19_415_680 seconds
Mars_T    = 686.98 * 24 * 3600      # 59_355_072 seconds
Jupiter_T = 4332.59 * 24 * 3600     # 374_335_776 seconds
Saturn_T  = 10759.22 * 24 * 3600    # 929_596_608 seconds
Uranus_T  = 30688.5 * 24 * 3600     # 2_652_836_800 seconds
Neptune_T = 60182 * 24 * 3600       # 5_199_244_800 seconds
Pluto_T   = 90560 * 24 * 3600       # 7_824_384_000 seconds
Eris_T    = 203830 * 24 * 3600      # 17_613_312_000 seconds (approx)



#IMPORTS

from __future__ import absolute_import, division, print_function
from dataclasses import dataclass
from typing import Tuple
from contextlib import contextmanager
from copy import deepcopy

import numpy as np
from numpy import atan2, floor, fmod, isnan
from numpy import cos, dot, sin, sqrt
from numpy.linalg import norm
#from math import atan2, floor, fmod, ising, isnan #chatqpt advice
from scipy.constants import pi


# Generator and context manager

@contextmanager
def saved_state(orbit):
    """Context manager to restore the orbit state upon leaving the block.
    
    This context manager saves the state of the orbit object before executing
    the code inside the block. After the block completes, the orbit object
    is restored to its original state.
    
    Args:
        orbit: The orbit object whose state is to be saved and restored.
    
    Yields:
        None: This function is used within a `with` block and doesn't return a value.
    """
    state = deepcopy(orbit.__dict__)
    yield
    orbit.__dict__.update(state)


def lookahead(collection, fillvalue=None):
    """Generates a series with lookahead to the next item in the collection.
    
    This function iterates over a collection and yields each item alongside the
    next item in the collection. The last item is paired with the `fillvalue` (default `None`).
    
    Args:
        collection: The collection (e.g., list or iterable) to iterate over.
        fillvalue: The value to pair with the last item in the collection if no next item exists.
    
    Yields:
        Tuple: A pair of the current and next items in the collection.
    """
    first = True
    for next_item in collection:
        if first:
            first = False
        else:
            yield current_item, next_item
        current_item = next_item
    yield current_item, fillvalue


# 3d vector class for easy shenanigans #gpt advice

@dataclass
class Vector3D:
    x: float
    y: float
    z: float

    @property
    def array(self):
        """Returns the vector as a NumPy array.
        
        This property converts the 3D vector into a NumPy array for easier
        manipulation with vectorized operations.
        
        Returns:
            np.array: The vector as a NumPy array [x, y, z].
        """
        return np.array([self.x, self.y, self.z])

    def cross(self, other):
        """Calculates the cross product of this vector with another vector.
        
        Args:
            other (Vector3D): The vector to compute the cross product with.
        
        Returns:
            Vector3D: The resulting vector from the cross product.
        """
        result = np.cross(self.array, other.array)
        return Vector3D(*result)

    def __add__(self, other):
        """Adds this vector with another vector.
        
        Args:
            other (Vector3D): The vector to add to this vector.
        
        Returns:
            Vector3D: The resulting vector after addition.
        """
        result = self.array + other.array
        return Vector3D(*result)

    def __sub__(self, other):
        """Subtracts another vector from this vector.
        
        Args:
            other (Vector3D): The vector to subtract from this vector.
        
        Returns:
            Vector3D: The resulting vector after subtraction.
        """
        result = self.array - other.array
        return Vector3D(*result)

    def __repr__(self):
        """String representation of the Vector3D object."""
        return f"{self.__class__.__name__}(x={self.x}, y={self.y}, z={self.z})"

# Position and Velocity types for clarity
@dataclass
class Position(Vector3D):
    """Represents the position vector in 3D space."""
    pass

@dataclass
class Velocity(Vector3D):
    """Represents the velocity vector in 3D space."""
    pass

# Simple state container
@dataclass
class StateVector:
    """Represents the state vector of an object, including its position and velocity.
    
    Attributes:
        position (Position): The position vector of the object.
        velocity (Velocity): The velocity vector of the object.
    """
    position: Position
    velocity: Velocity

# Orbital elements container
@dataclass
class OrbitalElements:
    """Represents the orbital elements of an orbiting object.
    
    Attributes:
        a (float): Semi-major axis.
        e (float): Eccentricity.
        i (float): Inclination.
        raan (float): Right ascension of the ascending node.
        arg_pe (float): Argument of periapsis.
        theta (float): True anomaly.
    """
    a: float        # semi-major axis
    e: float        # eccentricity
    i: float        # inclination
    raan: float     # right ascension of ascending node
    arg_pe: float   # argument of periapsis
    theta: float    # true anomaly

# Angular momentum computation
def angular_momentum(position: Position, velocity: Velocity) -> Vector3D:
    """Calculates the angular momentum vector from position and velocity vectors.
    
    Args:
        position (Position): The position vector.
        velocity (Velocity): The velocity vector.
    
    Returns:
        Vector3D: The resulting angular momentum vector.
    """
    return position.cross(velocity)




#Orbital element 

def orbit_radius(a, e, theta):
    '''
    Calc orbit radius
    '''
    r = a * (1 - e^2) / (1 + e * cos(theta))
    return r


def elements_4_apsides(apo_radius, peri_radius):
    """
    Calc orbital elements from apoapsis and periapsis radii
    """
    a = (apo_radius + peri_radius) / 2
    e = (apo_radius - peri_radius) / (apo_radius + peri_radius)
    return a, e


def create_orbit(i, raan, arg_pe, theta) -> Tuple[Vector3D, Vector3D, Vector3D]:
    """
    Create an orbit from the given elements
    """
    u = arg_pe + theta

    sin_u = sin(u)
    cos_u = cos(u)
    sin_raan = sin(raan)
    cos_raan = cos(raan)
    sin_i = sin(i)
    cos_i = cos(i)

    U = np.array(
        [cos_u * cos_raan - sin_u * sin_raan * cos_i,
         cos_u * sin_raan + sin_u * cos_raan * cos_i,
         sin_u * sin_i]
    )

    V = np.array(
        [-sin_u * cos_raan - cos_u * sin_raan * cos_i,
         -sin_u * sin_raan + cos_u * cos_raan * cos_i,
         cos_u * sin_i]
    )

    W = np.array(
        [sin_raan * sin_i,
         -cos_raan * sin_i,
         cos_i]
    )

    return U, V, W

def node_vector(angular_momentum: Vector3D) -> Vector3D:
    """Calculates the node vector from the angular momentum vector.
    
    The node vector is perpendicular to both the orbital plane and the angular momentum vector.
    
    Args:
        angular_momentum (Vector3D): The angular momentum vector.
    
    Returns:
        Vector3D: The node vector.
    """
    tmp = Vector3D(0, 0, 1)
    return angular_momentum.cross(tmp)

def eccentricity_vector(position: Position, velocity: Velocity, mu: float) -> Vector3D:
    """Calculates the eccentricity vector from position and velocity vectors.
    
    The eccentricity vector points from the center of the orbit toward periapsis.
    
    Args:
        position (Position): The position vector.
        velocity (Velocity): The velocity vector.
        mu (float): The standard gravitational parameter.
    
    Returns:
        Vector3D: The resulting eccentricity vector.
    """
    r = position.array
    v = velocity.array
    h = angular_momentum(position, velocity).array
    e = (np.cross(v, h) / mu) - (r / norm(r))
    return Vector3D(*e)

def specific_energy(position: Position, velocity: Velocity, mu: float) -> float:
    """Calculates the specific orbital energy of the object.
    
    Specific energy is the sum of kinetic and potential energy per unit mass.
    
    Args:
        position (Position): The position vector.
        velocity (Velocity): The velocity vector.
        mu (float): The standard gravitational parameter.
    
    Returns:
        float: The specific orbital energy.
    """
    r = norm(position.array)
    v = norm(velocity.array)
    return (v^2 / 2) - (mu / r)

def elements_from_state(mu: float, position: Position, velocity: Velocity) -> OrbitalElements:
    #erőssen krdőjeles, mert tabbal készült
    """Calculates the orbital elements from state vectors (position and velocity).
    
    Args:
        mu (float): The standard gravitational parameter.
        position (Position): The position vector.
        velocity (Velocity): The velocity vector.
    
    Returns:
        OrbitalElements: The calculated orbital elements (semi-major axis, eccentricity, etc.).
    """
    r = position.array
    v = velocity.array
    h = angular_momentum(position, velocity).array
    n = node_vector(h).array

    # Eccentricity vector
    e_vec = eccentricity_vector(position, velocity, mu).array
    e = norm(e_vec)

    # Semi-major axis
    energy = specific_energy(position, velocity, mu)
    a = -mu / (2 * energy)

    # Inclination
    i = acos(h[2] / norm(h))

    if abs(i - 0) < 1e-9:
        # For non-inclined orbits, raan is undefined;
        # set to zero by convention
        raan = 0
    if abs(e - 0) < 1e-9:
        # For circular orbits, place periapsis
        # at ascending node by convention
        arg_pe = 0
    else:
        # Argument of periapsis is the angle between
        # eccentricity vector and its x component.
        arg_pe = acos(e_vec.x / norm(e_vec))
    '''
        # Right ascension of ascending node
        raan = atan2(n[1], n[0])

        # Argument of periapsis
        arg_pe = atan2(e_vec[2], np.dot(n, e_vec))

        # True anomaly
        theta = atan2(np.dot(r, v), np.dot(r, e_vec))


    '''
    if abs(e - 0) < 1e-9:
        if abs(i - 0) < 1e-9:
            # True anomaly is angle between position
            # vector and its x component.
            f = acos(r.x / norm(r))
            if v.x > 0:
                f = 2 * pi - f
        else:
            # True anomaly is angle between node
            # vector and position vector.
            f = acos(dot(n, r) / (norm(n) * norm(r)))
            if dot(n, v) > 0:
                f = 2 * pi - f
    else:
        if e_vec.z < 0:
            arg_pe = 2 * pi - arg_pe

        # True anomaly is angle between eccentricity
        # vector and position vector.
        theta = acos(dot(e_vec, r) / (norm(e_vec) * norm(r)))

        if dot(r, v) < 0:
            theta = 2 * pi - theta
    return OrbitalElements(a=a, e=e, i=i, raan=raan, arg_pe=arg_pe, theta=theta)




##Conversions

def radius_from_alt(altitude, radius_body):
    """Converts altitude to radius from the center of the body.
    
    Args:
        altitude (float): The altitude above the surface.
        radius_body (float): The radius of the celestial body.
    
    Returns:
        float: The radius from the center of the body.
    """
    return radius_body + altitude

def alt_from_radius(radius, radius_body):
    """Converts radius from the center of the body to altitude.
    
    Args:
        radius (float): The radius from the center of the celestial body.
        radius_body (float): The radius of the celestial body.
    
    Returns:
        float: The altitude above the surface.
    """
    return radius - radius_body


def impulse_from_finite(acceleration, duration):
    """Returns the impulsive velocity change for a finite thrust burn.
    
    Args:
        acceleration (float): The acceleration produced by the thrust.
        duration (float): The duration of the thrust application.
    
    Returns:
        float: The impulsive velocity change.
    """
    return acceleration * duration






from setuptools import setup

setup(
    name='pymoodeng',
    version='0.99',
    description='A Python package for simple system simulations and potential transition fuel calculations.',
    author=['Kiss Márkus', 'Tari Zalán', 'Varga Benedek'],
    author_email=['kiss.markus@hallgato.ppke.hu', 'tari.zalan@itk.ppke.hu', 'varga.benedek@hallgato.ppke.hu'],
    packages=["numpy >= 2.2.4",
  "pandas >= 2.2.3" ,
  "matplotlib >= 3.10.1" ,
  "sphinx >= 8.2.3" ,
  "sphinx_rtd_theme >= 3.0.2"],
    install_requires=[
        'numpy',
        'pandas',
    ],
)